/* This file deals with IDE controller level stuff, as well as some utils
 * reguarding info on the disk
 */
#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/kmem.h>
#include <sys/ksynch.h>

#include <sys/dkio.h>
#include <sys/vtoc.h>

#include  <sys/dktp/fdisk.h>

#include <sys/ddi.h>
#include <sys/sunddi.h>

#include "defs.h"
#include "hpt-udma.h"

extern int debug_level;

/* Utils */
/* return REAL number of heads on disk, or 0 if not available */
int getnumheads(struct hpt_softstate *softstate,int drive){
	uint16_t *array=softstate->dev_info[drive];
	if(array==NULL) return -1;
	return array[3];
}
/* return REAL number of sectors per track, or 0 if not available */
int getnumsectpertrack(struct hpt_softstate *softstate,int drive){
	uint16_t *array=softstate->dev_info[drive];
	if(array==NULL) return -1;
	return array[6];
}
/* return full number of physical cylinders on the disk */
int getnumcyl(struct hpt_softstate *softstate,int drive){
	uint16_t *array=softstate->dev_info[drive];
	if(array==NULL) return -1;
	return array[1];
}

/************************************************************/
/* This section has a bunch of IDE register utilities       */
/************************************************************/

/* print to console the expanded version of the copy of the
 * status regisger passed in.
 * Since we only use printf(), it is now safe to call from interrupts
 */
void printstatus(uint8_t reg8){
	char buf1[50],buf2[50];
	buf1[0]='\0'; buf2[0]='\0';

	if(reg8&IDESTATUS_ERR){
		strcpy(buf2,buf1);
		sprintf(buf1,"%s ERR",buf2);
	}
	if(reg8&IDESTATUS_DATAREADY){
		strcpy(buf2,buf1);
		sprintf(buf1,"%s DATAREADY(DRQ)",buf2);
	}
	if(reg8&IDESTATUS_SERVICE){
		strcpy(buf2,buf1);
		sprintf(buf1,"%s SERVICE",buf2);
	}
	if(reg8&IDESTATUS_DF){
		strcpy(buf2,buf1);
		sprintf(buf1,"%s DF(fault)",buf2);
	}
	if(reg8&IDESTATUS_DEVREADY){
		strcpy(buf2,buf1);
		sprintf(buf1,"%s Dev_Ready",buf2);
	}
	if(reg8&IDESTATUS_BUSY){
		strcpy(buf2,buf1);
		sprintf(buf1,"%s BUSY",buf2);
	}
	printf("  print(ide)status: %s", buf1);

}

/* print to console, an interpreted version of err register bits.
 * Unfortunately, there arent a whole lot of different codes!
 */
void printideerr(uint8_t reg8){
	DLPRINT(1,"  printerr:");
	if(reg8&IDE_ERR_ABT){
		DLPRINT(1,"    Device ABRT");
	}
}

/* This reads the IDE status reguster. Note that if the BUSY bit is
 * set, all bets are off
 * Note that this is NOT ALWAYS HARMLESS.
 * Note also, that the mere fact of reading this register, sometimes
 * alters the state of things!
 *
 *  -> Odd as it may seem, if the UDMA engine is engaged, just READING
 *     this thing can BLOW UP THE SYSTEM!!!
 *
 *  -> If an interrupt is 'pending', reading this reg will CLEAR it.
 *
 * You probably want to use readaltstatus(), for most cases, unless
 * you are actually clearing the status of an operation.
 */
uint8_t readstatus(struct hpt_softstate *softstate,int bus){
	uint8_t reg8;

	reg8=ddi_get8(softstate->hpt_ide_command_handle[bus],
		      &softstate->ide_command_regs[bus]->command);
	return reg8;
}
/* This is a nice encapsulated way to read the IDE status register,
 * AND check to see if there is an error.
 * If an IDE command requires reading status register to finish,
 * call this INSTEAD OF readstatus()
 * return 1 on err, 0 on clear
 */
int got_ide_err(struct hpt_softstate *softstate,int bus){
	uint8_t reg8;
	reg8=readstatus(softstate,bus);
	if(reg8 & IDESTATUS_ERR){
		return 1;
	}
	return 0;
}

/* This reads the IDE ALTstatus register.
 * Unlike readstatus(), it should always be safe to call this...
 * except if the UDMA engine is busy.
 *
 * Note that if the BUSY bit is set, all bets are off, the bits
 * returned are most likely invalid.
 * And in fact... if the UDMA engine is busy.. even this read
 * can totally CRASH the >>>>ENTIRE SYSTEM<<<<
 */
uint8_t readaltstatus(struct hpt_softstate *softstate,int bus){
	uint8_t reg8;

	reg8=ddi_get8(softstate->hpt_ide_control_handle[bus],
		      &softstate->ide_control_reg[bus]->control);
	return reg8;
}

/*
 * There isnt actually a "clear_ide_err" command.
 * It is okay to just write a new command to the device, as long as
 * DRDY (device ready) is flagged.
 * send_idecmd() will automatically wait for DRDY.
 *
 * Be DURN SURE that you own the bus mutex, and the bus isnt busy,
 * etc, etc, etc
 */
uint8_t clear_ide_err(struct hpt_softstate *softstate,int bus){
	uint8_t reg8;
	reg8=readaltstatus(softstate,bus);
	return reg8;
}

/* checks to see if IDE stuff busy. DIFFERENT from got_udma_busy
 * reads ALT status register, to attempt to be nice.
 * However, if a UDMA operating is happening, this will still lock
 * up the machine, due to broken hardware 
 */
int got_ide_busy(struct hpt_softstate *softstate,int bus){
	uint8_t reg8;
	reg8=readaltstatus(softstate,bus);
	if(reg8 & IDESTATUS_BUSY){
		return 1;
	}
	return 0;
	
}


/* This reads the "error register". If the ERR bit is set in the
 * status register, this reg will tell you what KIND of err it is.
 * Use printideerr() to make it readable.
 */
uint8_t readerrval(struct hpt_softstate *softstate,int bus){
	uint8_t reg8;
	if(bus!=0){
		cmn_err(CE_WARN,"hpt: ERROR: we only readerrval() for bus0");
		return 0;
	}
	reg8=ddi_get8(softstate->hpt_ide_command_handle[bus],
		      &softstate->ide_command_regs[bus]->error);
	return reg8;
}


/* Yes, we WANT busy, sometimes */
/* This is a busywait waitforbusy :-)
 * return 0 if we got it, or 1 on timeout 
 * timeout is 1000 reads of the altstatus reg
 */
int waitforbusy(struct hpt_softstate *softstate, int bus){
	int count;

	for(count=0;count<1000;count++){
		if(got_ide_busy(softstate,bus)){
			return 0;
		}

	}

	return 1;
}

/* wait for "datarequest" flag. eg: "DATA READY,start transferring"
 * We loop using drv_usecwait, ignoring everything
 * but DATAREADY (DRQ) and ERR.
 *
 *  return 0 if got DRQ, 1 if ERR or timeout
 */
 int waitfordrq(struct hpt_softstate *softstate, int bus){
	uint8_t reg8;
	int count;

	reg8=readaltstatus(softstate,bus);

	for(count=0;count<1000;count++){
		if(reg8 & IDESTATUS_ERR)
			return 1;

		if(reg8&IDESTATUS_DATAREADY){
			return 0;
		}
		drv_usecwait(10);
		reg8=readaltstatus(softstate,bus);
	}
	return 1;
}


/* wait for "device ready" flag. eg: "DEV READY for command"
 * It also waits until busy is clear
 * So in theory, you can call this ONLY, instead of calling
 * wait_while_busy and then this routine
 *
 * return 0 ok , 1 if timeout
 */
int waitfordrdy(struct hpt_softstate *softstate,int bus){
	uint8_t reg8;
	int count;

	/* fast check */
	for(count=0;count<100;count++){
		reg8=readaltstatus(softstate,bus);
		if((reg8& (IDESTATUS_BUSY | IDESTATUS_DEVREADY))
		   == IDESTATUS_DEVREADY){
			return 0;
		}
	}

	for(count=0;count<2000;count++){
		reg8=readaltstatus(softstate,bus);
		if((reg8& (IDESTATUS_BUSY | IDESTATUS_DEVREADY))
		   == IDESTATUS_DEVREADY){
			return 0;
		}

		drv_usecwait(10);
	}

	/* Non-critical. We use this for "is drive there?" check as well
	 * as other things.
	 */
	if(debug_level>0){
		printf("End of waitfordrdy: failed: count=%d",count);
		printstatus(reg8);
	}

	return 1;
}

/* This is basically equivalent to a NOP, to waste a little time.
 * we read the 'altstatus' register, which should have NO side effect,
 * and throw the result away
 */
void read_delay(struct hpt_softstate *softstate,int bus){
	uint8_t reg8;
	reg8=readaltstatus(softstate,bus);
}

/* This is used to POLL, after we start a transfer, and the DRQ
 * bit has been set.
 * If DRQ gets stuck, return 1, else return 0
 *   (we use readstatus())
 */
 int wait_while_drq(struct hpt_softstate *softstate,int bus){
	int count;
	uint8_t reg8;

	/* yes, read ALT status */
	reg8=readstatus(softstate,bus);
	if((reg8 & IDESTATUS_DATAREADY)==0){
		return 0;
	}
	for(count=0;count<100;count++){
		reg8=readstatus(softstate,bus);
		if((reg8 & IDESTATUS_DATAREADY)==0){
			return 0;
		}
	}
	return 1;
}

/* If device is busy, wait until flag is clear. 
 * If BUSY gets stuck, return 1, else return 0
 *   (note: we use readaltstatus())
 * NEED TO NOT USE delay(), since we are called by waitfordata(),
 * in interrupt context.
 * 
 * XXX fix later with timeout or something?!!
 */
int wait_while_busy(struct hpt_softstate *softstate,int bus){
	int count;

	/* yes, read ALT status */
	if(!got_ide_busy(softstate,bus)){
		return 0;
	}

	/* yes, do it immediately again.
	 *  sometimes you just need the one cycle.
	 * Remember, using delay(1) is a context switch, and
	 * cannot be used from interrupt context
	 */

	if(!got_ide_busy(softstate,bus)){
		return 0;
	}
	/* First, try FAST wait. Then, try slow wait */
	for(count=1000;count>0;count--){
		if(!got_ide_busy(softstate,bus)){
			return 0;
		}
	}

	/* I find this disgustingly long. But if it is not this long,
	 * commands tend to flag as timed out when they
	 * will actually complete!
	 */
	for(count=2000;count>0;count--){
		drv_usecwait(100);
		if(!got_ide_busy(softstate,bus)){
			return 0;
		}
	}

	return 1;/* Still busy! */

}




/* After a drive reset, (or poweron, etc), a non-PACKET device is supposed
 * to have its registers set a specific way.
 * This checks for that "signature", and returns 0 if found, or 1 if not.
 */
int check_for_cleanide(struct hpt_softstate * softstate,int bus){
	uint8_t sc,sn,cl,ch,dh;
	sc=ddi_get8(softstate->hpt_ide_command_handle[bus],
		      &softstate->ide_command_regs[bus]->sector_count);

	sn=ddi_get8(softstate->hpt_ide_command_handle[bus],
		      &softstate->ide_command_regs[bus]->sector_number);

	cl=ddi_get8(softstate->hpt_ide_command_handle[bus],
		      &softstate->ide_command_regs[bus]->cylinder_low);
	ch=ddi_get8(softstate->hpt_ide_command_handle[bus],
		      &softstate->ide_command_regs[bus]->cylinder_high);
	dh=ddi_get8(softstate->hpt_ide_command_handle[bus],
		      &softstate->ide_command_regs[bus]->drive_head);

	if(debug_level>1){
		printf(" signature values: 0x%x,0x%x,0x%x,0x%x,0x%x",
		       sc,sn,cl,ch,dh);
	}
	if(sc!=1) return 1;
	if(sn!=1) return 1;
	if(cl!=0) return 1;
	if(ch!=0) return 1;
	if(dh!=0) {
		/*return 1;*/
		DLPRINT(1,"WARNING: ignoring bad signature on drive/head reg");
	}

	DLPRINT(1,"check_for_cleanide: got clean 'signature'");
	return 0;

}

/*
 * Use this routine to send a command to the IDE part of the chipset.
 * WE ASSUME YOU HAVE ALREADY SET DISK SELECT REGISTER APPROPRIATELY
 *
 * We also assume that the UDMA engine is off. otherwise, you may hang
 * the entire machine, on intel anyways.
 *
 * This will wait for the inevitable idestatusBUSY, when the device actually
 * starts decoding the command.
 * This uses ALTSTATUS, to check that BSY and DRQ are clear, before
 * writing the IDE reg command register.
 * If they are not clear, returns 1
 *
 * Also checks status AFTER writing command. If ERR flag detected,
 * returns 1
 *
 * We do actually try to WAIT OUT any BSY, subject to our timeout
 * in wait_while_busy()
 *
 * CONTEXT
 *  This routine now uses only printf(), not cmn_err. SO it is safe
 *  to be used from all contexts.
 *
 * RETURN 0 on okay, 1 on fail
 * If you're doing NOP, with no interrupt, call with check=0
 *
 * MUTEXES: You should probably hold bus_mutex before calling this.
 */
static int send_idecmd(struct hpt_softstate * softstate,int bus,uint8_t cmd,
		int check_status)
{
	uint8_t reg8;
	reg8=readaltstatus(softstate,bus);
	if(reg8 & IDESTATUS_DATAREADY){
		printf("hpt: send_idecmd: ERROR: IDE dataready flag set before cmd"
		 );

		printstatus(reg8);

		return 1;
	}
	if(reg8 & IDESTATUS_BUSY){
		if(wait_while_busy(softstate,bus) != 0){
			printf("hpt: send_idecmd: BUSY stuck");
			return 1;
		}
	}

	if(waitfordrdy(softstate,bus) != 0){
		/* we use IDENTIFY to see if disk is there, so dont complain*/
		if(cmd!=IDECMD_IDENTIFY){
			printf("hpt: send_idecmd: timeout on waitfordrdy");
		}
		return 1;
	}

	/* Write the actual command to the hardware!!!! */
	ddi_put8(softstate->hpt_ide_command_handle[bus],
		 &softstate->ide_command_regs[bus]->command, cmd);

	if(check_status==0){
		DLPRINT(2,"hpt: send_idecmd returning without reading status");
		return 0;
	}


	/* many times we are supposed to wait 400ns before reading
	 * the status register, after doing something.
	 * This is a recommended harmless way to do that.
	 */
	readaltstatus(softstate,bus);

	/* supposedly, some commands have the device 'block' read of
	 * this register until the dev is finished!
	 */
	if(got_ide_busy(softstate,bus)){
		if(wait_while_busy(softstate,bus) != 0){
			  printf("hpt: ERR: send_idecmd:  busied too long!");
			  return 1;
		}
		reg8=readaltstatus(softstate,bus);
	}

	if(reg8 & IDESTATUS_ERR){
		if(cmd==IDECMD_IDENTIFY){
			/* we half expect this particular command to fail.
			 * This is the way we detect if a drive is there
			 * So dont be verbose about it
			 */
			return 1;
		}
		printf("hpt:  send_idecmd(%x) hit ERR",cmd);

		printstatus(reg8);
		reg8=ddi_get8(softstate->hpt_ide_command_handle[bus],
			      &softstate->ide_command_regs[bus]->error);
		printf("hpt:  send_idecmd:  error reg is %x",reg8);

		/* we have to read the "real" status reg, to "clear" the
		 * error flag
		 */
		reg8=readstatus(softstate,bus);
		drv_usecwait(1);/* likewise, delay for "clear" purposes */
		return 1;
	}

	DLPRINT(2,"hpt: send_idecmd returning OK");
	return 0;
}
int idecmd_checkstat(struct hpt_softstate * softstate,int bus,uint8_t cmd){
	return send_idecmd(softstate,bus,cmd,1);
	
}
int idecmd_nocheck(struct hpt_softstate * softstate,int bus,uint8_t cmd){
	return send_idecmd(softstate,bus,cmd,0);
	
}

/* "bus" is superfluous, kinda. 
 * This is just a shortcut, to ease code reuse between hpt and bmi
 * note that the "control" register, is the SAME REGISTER as the
 * "altstatus" register. wierd, huh?
 */
void write_idecontrol(struct hpt_softstate * softstate,int bus, uint8_t val)
{
	ddi_put8(softstate->hpt_ide_control_handle[bus],
		 &softstate->ide_control_reg[bus]->control, val);
}

/* ONE step in most ide command setups.
 * Set the "drive/head" register for the drive you want to access.
 *
 * Note that we also always set LBA mode while we're doing it.
 *
 * This waits for the drive to both be not busy, AND for the "ready
 * bit to be shown.
 *
 * drive num can theoretically be 0-3, which determines bus
 * number as well.
 *
 * return 0 on okay, 1 on fail (BUSYied out or something)
 */
int ide_select_drive(struct hpt_softstate * softstate,int drive){
	uint8_t drivereg=LBA_BASE;
	int bus=0;
	if(drive>1) bus=1;

	if(drive & 0x01){
		drivereg|= HEAD_CHANNEL2;
	}

	/* waitfordrdy also checks busy bit */
	if(wait_while_busy(softstate,bus)!=0){
		printf("ERROR: ide_select_drive stuck busy");
		return 1;
	}

	/* I dont think we should wait for DRDY, since we are switching
	 * to a NEW device. Who cares if old device is ready?
	 */
	/*
	if(waitfordrdy(softstate,bus) != 0){
		printf("ERROR: ide_select_drive timeout waiting for drdy");
		return 1;
	}
	*/
	ddi_put8(softstate->hpt_ide_command_handle[bus],
		 &softstate->ide_command_regs[bus]->drive_head,
		 drivereg);

	readaltstatus(softstate,bus);/* delay */
	return 0;
}
/* reads drive/head register to find out which drive triggered interrupt*/
int ide_which_drive(struct hpt_softstate * softstate,int bus){
	uint8_t reg8;
	int drive=0;
	
	reg8=ddi_get8(softstate->hpt_ide_command_handle[bus],
		      &softstate->ide_command_regs[bus]->drive_head);
	if(reg8 & HEAD_CHANNEL2){
		drive=1;
	}
	return drive;
	
}

/* positiondisk() tells the ATA side of things, "here's where I want
 * to start the IO on the disk, which disk, AND here's how long on it."
 *  numblocks==0 actually means ***256** blocks!
 *
 * This is a separate function, so we can change our mind where to put this
 * routine ! :-) 
 * I attempt to use all 28 bits possible for LBA addressing.
 *
 * not sure all these wait_while_busys are needed.
 * But the FIRST one is, I'm pretty sure.
 * Definately the first (alt)status check is needed also
 *
 * we assume disk  is 0-3, so we deduce bus number
 *
 * MUTEXES:
 *	you should probably hold bus_mutex before calling here.
 */
int positiondisk(struct hpt_softstate * softstate, int disk,
    uint32_t startblock, uint8_t numblocks)
{

	uint8_t diskselect=LBA_BASE;
	uint8_t reg8;
	int bus=0;
	if(disk>1){
		bus=1;
	}

	DLPRINT2(2, "positiondisk called, for startblock %d",
		startblock);
	if((disk&0x01)==1){
		diskselect|=HEAD_CHANNEL2;
		DLPRINT(2,"positiondisk using CHANNEL 1?!");
	}

	reg8=readaltstatus(softstate,bus);
	if(reg8 & IDESTATUS_DATAREADY){
#ifndef  USE_SOFTINTR
		cmn_err(CE_WARN,"positiondisk entry: ERROR: DATA READY?!!");
		printstatus(readstatus(softstate,bus));
#endif
		return 1;
	}

	DLPRINT(2,"positiondisk setting up registers");
	if(wait_while_busy(softstate,bus) !=0){
		return 1;
	}
	ddi_put8(softstate->hpt_ide_command_handle[bus],
		 &softstate->ide_command_regs[bus]->sector_count,numblocks);

	if(wait_while_busy(softstate,bus) !=0){
		return 1;
	}

	/*
	 * In theory, LBA supports 28 bits. 
	 *
	 * remember, littleendian, so breaking up a sector-number ref
	 * has to go    0x00SSLLHH:
	 * S=sector, L=low-cyl, H=high-cyl
	 * Which means max of.. erm.. 8 gigs. oops.
	 * But the highest 4 bits go in dev/head register, which
	 * bumps us up to 128 gigs max
	 *
	 */
	reg8=(startblock)&0xff;
	ddi_put8(softstate->hpt_ide_command_handle[bus],
		 &softstate->ide_command_regs[bus]->sector_number,reg8);

	if(wait_while_busy(softstate,bus)!=0){
		return 1;
	}

	reg8=(startblock>>8)&0xff;
	ddi_put8(softstate->hpt_ide_command_handle[bus],
		 &softstate->ide_command_regs[bus]->cylinder_low,reg8);

	if(wait_while_busy(softstate,0)!=0){
		return 1;
	}

	reg8=(startblock>>16)&0xff;
	ddi_put8(softstate->hpt_ide_command_handle[bus],
		 &softstate->ide_command_regs[bus]->cylinder_high,reg8);
	if(wait_while_busy(softstate,bus)!=0){
		return 1;
	}


	diskselect|= (startblock>>24)&0x0f;  /* HIGHEST 24-27 bits! */

	ddi_put8(softstate->hpt_ide_command_handle[bus],
		 &softstate->ide_command_regs[bus]->drive_head, diskselect);

	if(wait_while_busy(softstate,bus)!=0){
		return 1;
	}

	reg8=ddi_get8(softstate->hpt_ide_command_handle[bus],
		 &softstate->ide_command_regs[bus]->sector_count);

	DLPRINT(2,"hpt: positiondisk completed");

	return 0;
}

/* I THINK I'm doing this right. arrg.
 * caller can call this IMMEDIATELY after calling
 * send_idecmd();
 *
 * We deliberately have a waitforbusy() here as the first thing.
 *
 * This reads the MAIN status register once, before proceeding
 * with PIO read!
 *
 * This reads a full 'block' of data.
 * So dest had better point to that much free memory!!!
 *
 * we assume disk  is 0-3, so we deduce bus number
 *
 * MUTEXES
 * You should have the bus mutex before calling.
 * Note that this polls for data, so interrupts should be disabled
 *
 * return 0 on okay, 1 on fail
 */
int getpiodatablock(struct hpt_softstate *softstate, int disk,
			    uint16_t *dest){
	int retval=0;
	uint8_t reg8;
	uint16_t reg16;
	int count=256; /* 2 bytes *256 == 512 == 1 block */
	int bus=0;
	if(disk>1){
		bus=1;
	}

	if(wait_while_busy(softstate,bus) != 0){
		DLPRINT(1,"hpt: getpiodatablock timed out in BUSY");
		printstatus(readaltstatus(softstate,bus));
		return 1;
	}
	if(waitfordrq(softstate,bus) != 0){
		DLPRINT(1,"hpt: getpiodatablock did not see DRQ");
		printstatus(readaltstatus(softstate,bus));
		return 1;
	}

	DLPRINT(2,"hpt: getpiodata saw DRQ, now getting data");
	for(count=256;count>0; count--){
		if(wait_while_busy(softstate,bus)!=0){
			DLPRINT(1,"hpt: getpiodatablock got stuck in BUSY");
			return 1;
		}

		reg8=readstatus(softstate,bus);
		if(reg8 & IDESTATUS_ERR){
			DLPRINT(1,"hpt: getpiodatablock detected ERR status");
			DLPRINT2(1, "   at count %d",count);
			return 1;
		}
		if((reg8 & IDESTATUS_DATAREADY) ==0){
			DLPRINT(1,"hpt: getpiodatablock no BUSY, but no DRQ either");
			return 1;
		}

	
		/* it's good to go! */
		/* Yes, the struct shows 'data' as only uint8_t.
		 * Just try not to think about it
		 */
		reg16=ddi_get16(softstate->hpt_ide_command_handle[bus],
		      (uint16_t*) &softstate->ide_command_regs[bus]->data);
		*dest++=reg16;
	}

	return retval;
}

/* putpiodatablock
 * should be called immediately after sending IDECMD_WRITE.
 * It will read the main IDE status register to show we want a response
 * from the drive.
 * Writes exactly 1 block from the source pointer
 * disk may be 0-3, and we'll figure out the correct bus.
 *
 * You should have the bus mutex before calling.
 * we will not set or release it here.
 *
 * return 0 on okay, 1 on fail
 */
int putpiodatablock(struct hpt_softstate *softstate, int disk,
			    uint16_t *source){
	int retval=0;
	uint8_t reg8;
	int count=256; /* 2 bytes *256 == 512 == 1 block */
	int bus=0;
	if(disk>1){
		bus=1;
	}

	if(waitfordrq(softstate,bus) != 0){
		DLPRINT(1,"hpt: putpiodatablock did not see DRQ");
		printstatus(readaltstatus(softstate,bus));
		return 1;
	}

	DLPRINT(2,"hpt: putpiodata saw DRQ, now writing data");

	for(count=256;count>0; count--){
		/* Hmmm. Other drivers dont check for wait between
		   writes: they just slam the whole block in.
		   (eg use ddi_rep_put16)
		   But I'll be paranoid for now.
		 */
		if(wait_while_busy(softstate,bus)!=0){
			DLPRINT(1,"hpt: putpiodatablock got stuck in BUSY");
			return 1;
		}

		reg8=readstatus(softstate,bus);
		if(reg8 & IDESTATUS_ERR){
			DLPRINT(1,"hpt: putpiodatablock detected ERR status");
			DLPRINT2(1, "   at count %d",count);
			return 1;
		}
		if((reg8 & IDESTATUS_DATAREADY) ==0){
			DLPRINT(1,"hpt: putpiodatablock no BUSY, but no DRQ either");
			return 1;
		}

	
		/* It's good to go! */
		/* Yes, the struct shows 'data' as only uint8_t.
		 * Just try not to think about it
		 */
		ddi_put16(softstate->hpt_ide_command_handle[bus],
		      (uint16_t*) &softstate->ide_command_regs[bus]->data,
		      *source++);
	}

	/* If we dont wait until clear here, then
	 * caller might enable interrupts, and the disk will trigger an
	 * interrupt when it finishes the transfer
	 */
	/* dummy delay */
	reg8=readaltstatus(softstate,bus);
	/* and an extra, paranoid delay */
	drv_usecwait(10);
	
	if(wait_while_busy(softstate,bus)!=0){
		DLPRINT(1,"putpiodatablock: BUSY did not clear after transfer");
		retval=1;
	}
	DLPRINT(2,"putpiodata: BUSY flag now clear. returning OK.");

	return retval;
}


/* DO NOT USE THIS unless you already understand it.
 * This is to check timings of the various phases of
 * putting an IDE command to the drive.
 */
int debug_idecmd(struct hpt_softstate * softstate,int bus,uint8_t cmd){
	uint8_t reg8;
	
	reg8=readaltstatus(softstate,bus);

/*	still need to do this?*/
	return 0;
}

/* This should only be called in kernel context, in attach() */
/* This does absolutely NO LOCKING */
void enable_readaread(struct hpt_softstate * softstate,int drive){
	uint8_t feature=IDE_FEAT_READAHEAD;
	uint8_t command=IDECMD_SETFEATURES;
	uint8_t reg8;
	int bus=0; /* for b.m.i., this should be fixed */

	if(drive>1) bus=1;
	if(drive>HPT_NUMDRIVES){
		cmn_err(CE_WARN,"Error: drivenum too high in enable_readahead");
		return;
	}

	if(ide_select_drive(softstate,bus)!=0){
		printf("Error selecting drive for enable_readahead");
		return;
	}

	ddi_put8(softstate->hpt_ide_command_handle[bus],
		 &softstate->ide_command_regs[bus]->error, feature);
	/* delay */
	reg8=readaltstatus(softstate,bus);

	ddi_put8(softstate->hpt_ide_command_handle[bus],
		 &softstate->ide_command_regs[bus]->command, command);

	if(wait_while_busy(softstate,bus)!=0){
		cmn_err(CE_WARN,"hpt: enable_readahead timeout after command?");
		return;
	}

	reg8=readstatus(softstate,0);

	if(reg8 & IDESTATUS_ERR){
		cmn_err(CE_WARN,"hpt: enable_readahead command ERRed?");
		return;
	}
	DLPRINT2(1,"hpt: set readahead on drive %d",drive);

}

void disable_interrupt(struct hpt_softstate * softstate, int bus){
	uint8_t reg8;
	DLPRINT(1,"Disabling interrupts");
	write_idecontrol(softstate,bus,IDECONTROL_NOIRQ);

	drv_usecwait(100);

	/* this read will hopefully clear the status bits */
	reg8=readstatus(softstate,bus);
}

/* Note that this will set the channel # to 0 */

void enable_interrupt(struct hpt_softstate * softstate,int bus){
	uint8_t reg8;
	reg8=0;
	DLPRINT(1,"Enabling interrupts");
	write_idecontrol(softstate,bus,reg8);

	/* this is to hopefully let the IRQ flag clear */
	drv_usecwait(10);

	/* this read actually clears the status bits */
	reg8=readstatus(softstate,bus);
}




/*
 * But for reading stuff in the solaris sector before a label is written,
 *  'format' thinks it needs this command.
 * Unfortunately, its too much of a pain to randomly insert things in the
 *  DMA based readwrite queue. It keeps crashing the system when
 *  trying to free the dma handle)
 *
 *  Note that sector numbers in this function are RAW DISK sectors.
 *  sector0==first block on disk
 *
 *   note: "sector" == "block"
 *
 * MUTEXES and interrupts
 *   This will grab and release the bus mutex
 *   It polls for data, using getpiodatablock()
 *   YOU MUST HAVE DISABLED INTERRUPTS before using this!!!
 *
 * return 1 on error, 0 on okay
 */

int hpt_read_pio(struct hpt_softstate * softstate,int disk,
	     	void *bufferp,int sectornum, uint8_t sectorcount)
{
	uint8_t *sectorbuf=bufferp;
	int bus=0;
	int retval=0;

	if(disk>1){
		bus=1;
	}
	mutex_enter(&softstate->bus_mutex[bus]);


	do {
		if(positiondisk(softstate,disk,sectornum,sectorcount) !=0){
			retval=1; break;
		}

		if(idecmd_checkstat(softstate,bus,IDECMD_READ)!=0){
			cmn_err(CE_WARN,"Error starting PIO IDE READ ");
			retval=1; break;
		}

		while(sectorcount>0){
			if(getpiodatablock(softstate,disk,bufferp) !=0){
				cmn_err(CE_WARN,"Error reading pio data");
				retval=1; break;
				}
			sectorcount--;
			sectorbuf+= HPT_SECTSIZE;
		}
	} while(0);

	/* This is NOT ACTUALLY REQUIRED by the ATA spec. There should
	 * be no pending interrupt for the PIO read sequence.
	 * But I want to make this match the write routine code.
	 */
	readstatus(softstate,bus);
	
	mutex_exit(&softstate->bus_mutex[bus]);
	return retval;
}

/*
 * write via "PIO" (not DMA) method, to disk.
 *
 * MUTEXES:
 * This will grab the bus mutex!!!
 *
 * This does things by hand, so DISABLE INTERRUPTS before calling
 *
 * returns 0 on okay, 1 on IO error
 */
int hpt_write_pio(struct hpt_softstate * softstate,int disk,
	     	void *bufferp,int sectornum, uint8_t sectorcount)
{
	uint8_t *sectorbuf=bufferp;
	int bus=0;
	int retval=0;

	if(disk>1){
		bus=1;
	}
	mutex_enter(&softstate->bus_mutex[bus]);

	do {
		if(positiondisk(softstate,disk,sectornum,sectorcount) !=0){
			retval=1; break;
		}
		if(idecmd_checkstat(softstate,bus,IDECMD_WRITE)!=0){
			cmn_err(CE_WARN,"Error starting PIO IDE WRITE ");
			retval=1; break;
		}

		while(sectorcount>0){
			if(putpiodatablock(softstate,disk,bufferp) !=0){
				cmn_err(CE_WARN,"Error writing PIO data");
				retval=1; break;
				}
			sectorcount--;
			sectorbuf+= HPT_SECTSIZE;
		}
	} while(0);

	/* to hopefully clear the pending interrupt */
	readstatus(softstate,bus);
	
	mutex_exit(&softstate->bus_mutex[bus]);
	return retval;
}

/* software reset of whole bus (?)
 * It would be a good idea to make REALLY REALLY sure nothing else is
 *   writing to the bus right now!
 *
 * Note also, that you probably want to call init_udma on the
 * devices on the bus, which will then trigger the interrupt handler,
 * and clean up any pending stuff for it.
 *
 * return 0 on okay, 1 on fail
 */
int software_reset(struct hpt_softstate *softstate, int bus){
	uint8_t cmd=IDECONTROL_SRESET;
	uint8_t reg8;

	/* Because the UDMA engine is wierd about ATA access when busy,
	 * (locks up machine!) we have to first turn off any active job
	 */
	stop_udma(softstate,bus);
	
/* I'm attempting to follow section 9.2: Software reset protocol  (p234)
 * from the T13 spec. This is DIFFERENT FROM 9.11: device reset
 */
	if(debug_level>1){
		printf(" start of software_reset: ide status check");
		printstatus(readaltstatus(softstate,bus));		
	}


	if(got_ide_busy(softstate,bus)){
		DLPRINT(1,"warning: device BUSY before software reset. Waiting");
		if(wait_while_busy(softstate,bus) != 0){
			DLPRINT(1,"STILL busy. resetting anyway");
		}
	}


	write_idecontrol(softstate,bus,cmd);

	/* have to hold high for "at least 5 us."*/
	drv_usecwait(100);

	cmd=0;
	write_idecontrol(softstate,bus,cmd);

	/* have to leave as 0 for "at least 2 ms."*/
	drv_usecwait(2000);


	reg8=readstatus(softstate,bus); /* This finishes out the reset */
					/* reading normal status is required*/


	if(got_ide_err(softstate,bus)){
		DLPRINT(0,"software_reset: ide ERR set after device reset. Ignoring.");
		printstatus(readaltstatus(softstate,bus));		
	} else {
		DLPRINT(1,"software_reset: Reset happened: ide statusreg ERR-free");
	}

	/* And need to wait even longer, if BSY bit set */

	if(wait_while_busy(softstate,bus) != 0){
		DLPRINT(0,"device stuck BUSY after reset");
		return 1;
	}

	if(check_for_cleanide(softstate,bus)==1){
		printf("WARNING: no 'clean' IDE signature after software_reset!!");
	}

	if(debug_level>0){
		printf("AFTER  software_reset:");
		printstatus(readaltstatus(softstate,bus));
	}

	return 0;
	
}

/*
 * This resets specified drivenum. But you also have to wait until the
 * other drive on the bus is free before calling this.
 * (eg: This will either stop anything else in progress, or just fail)
 *
 * WARNING... If you are in some kind of error state, and are trying to
 * clear it by calling this routine... you better make sure the UDMA
 * stuff is cleared FIRST. Otherwise, you can hard lock the machine,
 * get bad filesystem corruption (even on OTHER disks!!!), ....
 *
 *
 * NOTE: this currently seems to trigger the ERR flag. which I figure
 * is an indication that I'm doing SOMTHING WRONG in here.
 *  {Hmmm.. namely, t13/1321D, section 6.4.1 says DEVRESET only valid
 *   for devices supporting PACKET command set?!!}
 *
 * Use software_reset() intead.
 *
 * returns 0 on okay, 1 on fail
 */
int reset_drive(struct hpt_softstate *softstate, int drivenum){
	uint8_t cmd=IDECMD_DEVRESET;
	uint8_t reg8=0,devselect;
	int bus=0;

/* I'm attempting to follow section 9.11: Device reset protocol 
 * from the T13 spec. This is DIFFERENT FROM 9.2: software reset
 */

	if(drivenum>1){
		DLPRINT(0,"reset_drive: error: this driver cannot have drivenum >1");
		return 1;
	}

	if(debug_level>1){
		printf("reset_drive: start: ide status check");
		printstatus(readaltstatus(softstate,bus));		
	}

	/* doing this manually, because ide_select_drive waits for
	 * non-busy
	 */
	devselect=LBA_BASE;
	if((drivenum&0x01)==1){
		devselect |=HEAD_CHANNEL2;
	}

	/* select drive in drive/head register */
	ddi_put8(softstate->hpt_ide_command_handle[bus],
		 &softstate->ide_command_regs[bus]->drive_head,
		devselect);

	if(got_ide_busy(softstate,bus)){
		DLPRINT(0,"warning: device BUSY before device reset. Waiting");
		if(wait_while_busy(softstate,bus) != 0){
			DLPRINT(1,"hpt: device STILL busy. resetting anyway");
		}
	}

	/* delay */
	reg8=readaltstatus(softstate,bus);

 
	/* "real" status */
	reg8=ddi_get8(softstate->hpt_ide_command_handle[bus],
			&softstate->ide_command_regs[bus]->command);

	DLPRINT(2,"reset_drive: doing actual ide command now");

	/* Want to force this write. Dont use idecmd_checkstat()..
	 */
	ddi_put8(softstate->hpt_ide_command_handle[bus],
		 &softstate->ide_command_regs[bus]->command, cmd);

	/* have to hold high for "at least 400 ns."*/
	drv_usecwait(10);
	/* delay */
	reg8=readaltstatus(softstate,bus);

	if(got_ide_err(softstate,bus)){
		DLPRINT(0,"drive_reset: ide ERR set after device reset. Ignoring.");
		printstatus(readaltstatus(softstate,bus));		
	} else {
		DLPRINT(1,"Drive Reset complete: ide statusreg is ERR-free");
	}

	/* And need to wait even longer, if BSY bit set */
	if(wait_while_busy(softstate,bus) != 0){
		DLPRINT(0,"device stuck BUSY after reset");
		return 1;
	}

	if(check_for_cleanide(softstate,bus)==1){
		DLPRINT2(1,"drive_reset: WARNING: no 'clean' IDE signature after reset, for drive %d",
			 drivenum);
	}
	return 0;
}

