/* This file holds routines that deal with fdisk and solaris
 * partitioning details
 */

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/file.h> /* for "FREAD" in ddi_copyout*/
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/kmem.h>
#include <sys/ksynch.h>

#include <sys/dkio.h>
#include <sys/vtoc.h>


#include <sys/dktp/fdisk.h>

#include <sys/ddidmareq.h>

/* THESE MUST BE LAST, of the sys includes */
#include <sys/ddi.h>
#include <sys/sunddi.h>

#include "defs.h"
#include "hpt.h"
#include "hpt-ide.h"
#include "hpt-partitions.h"

/**********************************************************************/
/* Here's the section for partitions, fdisk, and all that garbage     */
/**********************************************************************/


/* make "minor_node" dev files for all slices,
 * plus one big one for the entire disk,
 * Plus 4 for the primary partitions
 */
void make_minors(dev_info_t *dip, int instance, int disknum){
	int minor=instance*DEV_INSTANCEMULT;
	char devname[20];
	int slice,fpart;

	sprintf(devname,"hpt%dD%d",instance,disknum);
	if(ddi_create_minor_node(dip, devname, S_IFBLK,
			 minor|HPT_FULLDISK, "UDMA", 0) != DDI_SUCCESS){
		cmn_err(CE_WARN,"could not make 'whole drive' device node");
	}
	sprintf(devname,"hpt%dD%d,raw",instance,disknum);
	if(ddi_create_minor_node(dip, devname, S_IFCHR,
			 minor|HPT_FULLDISK, "UDMA", 0) != DDI_SUCCESS){
		cmn_err(CE_WARN,"could not make 'whole drive' device node");
	}


	/* "fdisk partition" minor nodes */
	for(fpart=1;fpart<5; fpart++){
		sprintf(devname,"hpt%dD%df%d",instance,disknum,fpart);
		if(ddi_create_minor_node(dip, devname, S_IFBLK,
					 minor|HPT_FULLDISK|fpart,
					 "UDMA", 0)  != DDI_SUCCESS)
		{
			cmn_err(CE_WARN,"partition dev node creation failed");
		}
		sprintf(devname,"hpt%dD%df%d,raw",instance,disknum,fpart);
		if(ddi_create_minor_node(dip, devname, S_IFCHR,
					 minor|HPT_FULLDISK|fpart,
					 "UDMA", 0)  != DDI_SUCCESS)
		{
			cmn_err(CE_WARN,"partition dev node creation failed");
		}
	}

	/* "solaris slice" minor nodes */
	for(slice=0;slice<16;slice++){
		sprintf(devname,"hpt%dd%d%c",instance,disknum,'a'+slice);
		if(ddi_create_minor_node(dip, devname, S_IFBLK,
					 minor+slice,
					 "UDMA", 0)  != DDI_SUCCESS)
		{
			cmn_err(CE_WARN,"partition dev node creation failed");
			return;
		}
		sprintf(devname,"hpt%dd%d%c,raw",instance,disknum,'a'+slice);
		if(ddi_create_minor_node(dip, devname, S_IFCHR,
					 minor+slice, "UDMA", 0)
		   != DDI_SUCCESS)
		{
			cmn_err(CE_WARN,"raw device node creation failed");
			return;
		}
	}
	
}

/* Number of "sectors" the drive can hold 
 * dev_info[drive][58,57] is traditional sector count
 * 
 * (you could generate numblocks by multipling by blockspersector)
 *
 * dev_info[drive][61,60] is LBA sector count, where
 * 1 sector==1 block, i THINK.
 */
uint32_t drivenumsect(struct hpt_softstate*softstate,int drive){
	uint32_t numsect=(softstate->dev_info[drive][58]<<16);
	numsect|=softstate->dev_info[drive][57];

	return numsect;
}

/* Return total number of blocks in a drive.
 * This is ENTIRE DRIVE. Not just the solaris fdisk partition.
 * Note that technically, this is not the same as drivenumsect,
 * although if sectorsize==blksize, they SHOULD return the
 * same number. Except that from the drive's perspective, they
 * probably are not the same number
 */
uint32_t drivenumblk(struct hpt_softstate*softstate,int drive){
	uint32_t numblk=(softstate->dev_info[drive][61]<<16);
	numblk|=softstate->dev_info[drive][60];

	return numblk;
}

/* just like getfstartblk, except without the conversion. 
 * Returns ABSOLUTE (physical) BLOCK NUMBER of fdisk partition,
 * where block #0==fdisk info, and normally usable space
 * starts at block 1
 * 'fpart' indexing scheme starts at **1**,
 * unless you want the whole disk.
 * (this seemingly pointless case allowed for, so we dont have
 *   to special-case higher up)
 *
 * Ummm... Since we dont really convert to true disk 'sectors',
 * you should probably always use getfstartblk instead
 */
uint32_t getfstartsect(struct hpt_softstate * softstate,int disk,int fpart){
	uint32_t startsect;
	struct ipart *partinfo;

	if(fpart==0) return 0;
	if(fpart>FD_NUMPART) fpart=FD_NUMPART;

	partinfo=softstate->fdisklabel[disk];
	startsect=partinfo[fpart-1].relsect;
	return startsect;
}

/* translate relative fdisk partition sector start, to 
 *  (512byte) BLOCK address, relative to whole disk.
 * 'fpart' index starts at **1**, unless you want the whole disk.
 * (this seemingly pointless case allowed for, so we dont have
 *   to special-case higher up)
 */
uint32_t getfstartblk(struct hpt_softstate * softstate,int disk,int fpart){
	uint32_t startsect,startblk;
	struct ipart *partinfo;

	if(fpart==0) return 0;
	if(fpart>FD_NUMPART) fpart=FD_NUMPART;

	partinfo=softstate->fdisklabel[disk];
	startsect=partinfo[fpart-1].relsect;
	/*startblk=startsect * softstate->blkspersect[disk];*/
	startblk=startsect;

	return startblk;
}

/* translate size of fdisk in sectors, to size in blocks
 * (512byte) BLOCK
 * Count index as starting at **1**: specify whole disk as 0
 *
 * Note that the REAL ARRAY INDEX starts at 0 for "first" partition
 */
uint32_t getfblkcount(struct hpt_softstate * softstate,int disk,int fpart){
	uint32_t sectcount,blkcount;
	struct ipart *partinfo;

	if(fpart==0) return drivenumblk(softstate,disk);

	partinfo=softstate->fdisklabel[disk];
	sectcount=partinfo[fpart-1].numsect;
	/*blkcount=sectcount * softstate->blkspersect[disk];*/
	blkcount=sectcount;

	return blkcount;
}

/* Give total number of blocks in a solaris partition.
 * If vtoc invalid, you should get zeros back, due to way info is
 * initialized
 */
uint32_t getsblkcount(struct hpt_softstate * softstate,int disk,int partnum){
	uint32_t bcount;

	if(partnum>=FD_NUMPART)
		partnum=FD_NUMPART-1;

	bcount=softstate->partitions[disk].v_part[partnum].p_size;
	return bcount;
}

/* print 'fdisk partition(s)' information, of particular disk */
void print_fparts(struct hpt_softstate * softstate,int disk){
	int fpart;
	struct ipart *partinfo = softstate->fdisklabel[disk];

	cmn_err(CE_CONT, "	   relsect    "
			"numsect	 sysid	     bootid\n");
	for (fpart = 0; fpart < FD_NUMPART; fpart++) {
		cmn_err(CE_CONT,
			"    %d:  %d	%d	0x%x	   0x%x\n",
			fpart+1,
			(int)partinfo[fpart].relsect,
			(int)partinfo[fpart].numsect,
			(int)partinfo[fpart].systid,
			(int)partinfo[fpart].bootid);
	}

}

/* print solaris partition ' information, of particular disk,
 * that we have cached in our "softstate" info
 */
int print_sparts(struct hpt_softstate * softstate,int disk){
	int count;
	struct vtoc *vtocptr=&softstate->partitions[disk];

#ifdef SANECHECK
	if(vtocptr->v_sanity!= VTOC_SANE){
		cmn_err(CE_WARN,"solaris partition for disk %d not sane",
			disk);
		return 1;
	}
#endif
	DLPRINT2(1,"hpt: print_spart(): number of solaris slices=%d",
		vtocptr->v_nparts);

	cmn_err(CE_CONT, "	   relsect	numsect	 type\n");

	for (count = 0; count < V_NUMPAR; count++) {
		cmn_err(CE_CONT,
			"    %d:  %d	%d	%d\n",
		count,
		vtocptr->v_part[count].p_start,
		vtocptr->v_part[count].p_size,
		vtocptr->v_part[count].p_tag);
	}
	return 0;
}


/* hpt_auto_vtoc:
 * "normally", format does some scsi magic sense thing, to get information
 * to auto-label
 * But for some stupid reason, it doesnt do what it could do, for IDE disks.
 * So we have to auto-label.
 * SUN actually has a special driver just for this kind of junk, called
 * snlb. But it is undocumented, so I'll just do it myself.
 *
 * Call with the fdisk index of the solaris partition.
 * Be sure to use the same indexing scheme that getfblkcount() uses
 * for the fdisk partitions.
 *
 * Calculate the info, and copy it to the in-memory version we hold.
 *
 * THIS WILL BLOW AWAY ANY PREVIOUS in-memory version, so be sure you
 * want to call this.
 */

void hpt_auto_vtoc(struct hpt_softstate * softstate,int drive,
		   int fpartnum){

	struct vtoc *vtocptr=&softstate->partitions[drive];
	int nheads,nsect, blkspercyl;
	char volumelabel[LEN_DKL_ASCII];
	char name_array[21];
	int blksinfpart=0;
	uint16_t *id_array=softstate->dev_info[drive];

	hpt_dname_to_string(id_array,name_array);

	bzero(vtocptr,sizeof(struct vtoc));

	blksinfpart=getfblkcount(softstate,drive, fpartnum);


	nheads=getnumheads(softstate,drive);
	nsect=getnumsectpertrack(softstate,drive);
	blkspercyl=nheads*nsect;

	sprintf(volumelabel,"%s cyl %d alt 2 hd %d sec %d",
		name_array,
		blksinfpart/blkspercyl,
		nheads,nsect);

	/* really stupid useless "volume" labeling */
	/* It's exactly LEN_DKL_VVOL==8 chars long */
	sprintf(vtocptr->v_volume,"solaris");

	/* This is the "real" label, that 'format' actually uses */
	bcopy(volumelabel, vtocptr->v_asciilabel, LEN_DKL_ASCII);

	/* Now we invent sizes for the solaris partions.
	* We only make a "backup" slice2, and the other special
	* intel-required slice 8 and 9
	*/

	vtocptr->v_part[2].p_start=0;
	vtocptr->v_part[2].p_size=blksinfpart;
	vtocptr->v_part[2].p_tag=V_BACKUP;
	vtocptr->v_part[2].p_flag=V_UNMNT;

	/* fdisk partition bootblock holder */
	vtocptr->v_part[8].p_start=0;
	vtocptr->v_part[8].p_size=blkspercyl;
	vtocptr->v_part[8].p_tag=V_BOOT;
	vtocptr->v_part[8].p_flag=V_UNMNT;

	/* "alt sectors" tagged here.
	 * ONE AND ONLY ONE (throwaway) partition should hold this.
	 * Very cool. 'format' DOES USE THIS, if tagged!!
	 */
	vtocptr->v_part[9].p_start=blkspercyl;
	vtocptr->v_part[9].p_size=2*blkspercyl;
	vtocptr->v_part[9].p_tag=V_ALTSCTR;


	vtocptr->v_sectorsz=DEV_BSIZE;
	vtocptr->v_nparts=V_NUMPAR;
	vtocptr->v_version=V_VERSION;
	vtocptr->v_sanity=VTOC_SANE;
}

/* This routine copies the given "mboot" info into our
 * in-memory holder.
 * It then uses that info to read in a VTOC, if there is a solaris
 * partition.
 * If there is a solaris partition, but no valid VTOC, it will make
 * one up.
 *
 * Returns 0 on okay, 1 on FAIL
 */
int copy_fdisk_info(struct hpt_softstate*softstate,struct mboot *mbootcopy,int disk){
	struct ipart *fpartinfo;
	uint16_t *bufferp;
	int fcount, startblk;
	int bus=0;
	if(disk>1) bus=1;
	
	fpartinfo=softstate->fdisklabel[disk];
	bcopy(mbootcopy->parts, fpartinfo, sizeof(struct ipart) *FD_NUMPART);

	if(debug_level>0)
		print_fparts(softstate,disk);

	/* find the solaris partition, and update our info from it
	 *
	 * Note that the vtoc is NOT at blk0, but blk 1.
	 */

	for (fcount = 0; fcount < FD_NUMPART; fcount++) {

		if(fpartinfo[fcount].systid == SUNIXOS)
			break;
	}
	
	if(fpartinfo[fcount].systid != SUNIXOS){
		cmn_err(CE_NOTE, "warning: no solaris fdisk partition");
		return 0;
	}

	startblk=fpartinfo[fcount].relsect;

	/* sol_startblk notes the blocknumber we add to
	 * the SLICE starting block [p_start], to get its
	 * true position on the disk.
	 * It's a quickreference to
	 * "Where is block 0 as far as solaris is concerned?"
	 */
	softstate->sol_startblk[disk]=startblk;


	/* Remember, VTOC is at blk 1, NOT blk0!! */
	DLPRINT3(1,"trying to get solaris VTOC for fpart %d,disksec %d",
		fcount,	startblk+1);

	bufferp=kmem_alloc(HPT_SECTSIZE,KM_SLEEP);

	if(hpt_read_pio(softstate,disk,bufferp, startblk+1,1) !=0){
		cmn_err(CE_WARN,"IO error reading VTOC!");
		return ENXIO;
	}
	bcopy(bufferp, &softstate->partitions[disk],
				sizeof(struct vtoc));

	if(softstate->partitions[disk].v_sanity != VTOC_SANE){
		DLPRINT(1,"Bad Vtoc found: making one up");
		hpt_auto_vtoc(softstate,disk,fcount+1);
	} else {
		DLPRINT(2,"Good Vtoc found. Keeping in memory");
	}

	if(debug_level>1){
		int ibuf[10];
		bcopy(bufferp,ibuf,sizeof(int) *10);
		cmn_err(CE_NOTE,
		  " [RAW view of vtoc block@%d =%x:%x:%x:%x:%x:%x:%x:%x]",
			startblk,
			ibuf[0],ibuf[1],
			ibuf[2],ibuf[3],
			ibuf[4],ibuf[5],
			ibuf[6],ibuf[7]
			);
	}

	kmem_free(bufferp,HPT_SECTSIZE);
	if(debug_level>1){
		print_sparts(softstate,disk);
	}

	return 0;
}

/* Find the fdisk partition that is the solaris partition.
 * return -1 if none found, otherwise return FIRST MATCH to
 * fdisk type=SUNIXOS (0x82)
 */
int hpt_find_sol_fpart(struct hpt_softstate *softstate,int disk){
	int pnum;
	struct ipart *fdiskinfo;
	for(pnum=0;pnum<FD_NUMPART;pnum++){
		fdiskinfo=&softstate->fdisklabel[disk][pnum];
		if(fdiskinfo->systid == SUNIXOS)
			return pnum;
	}
	return -1;
}


/* Read blk0, which is the fdisk partitioning info.
 * We then call copy_fdisk_info, which also looks for a solaris partition,
 * and fills out our solaris partition info also.
 *
 * This should be called the first time a disk is attached. It can
 * also be called at any time. we will then update the partitioning
 * information from what's on the disk
 * Let the user beware. We allow this at ANY TIME.
 *
 * return 0 on good, 1 on fail
 */
int get_fdiskpart(struct hpt_softstate * softstate,int disk){
	struct mboot mbootcopy;
	uint16_t *bufferp;
	int bus=0;


	if(disk>1){
		bus=1;
	}
	
	if(softstate->dev_info[disk][0]==0){
		DLPRINT(1,"no dev_info. get_fdiskpart assumes invalid disk ");
		return 1;
	}

	bufferp=kmem_alloc(HPT_SECTSIZE,KM_SLEEP);

	if(hpt_read_pio(softstate,disk,bufferp,0,1)!=0){
		cmn_err(CE_WARN,"read of fdisk info failed");
		kmem_free(bufferp,HPT_SECTSIZE);
		return 1;
	}


	/* Supposedly, there are some alignment issues here. Blah */
	bcopy(bufferp, &mbootcopy, sizeof(mbootcopy));

	kmem_free(bufferp,HPT_SECTSIZE);

	if(mbootcopy.signature != MBB_MAGIC){
		DLPRINT(1,"Disk is not fdisk labeled!");
		return 1;
	}

	if(copy_fdisk_info(softstate,&mbootcopy,disk)==0){
		DLPRINT(1,"get_fdiskpart SUCCEEDED");
		return 0;
	} else {
		DLPRINT(1,"get_fdiskpart FAILED");
		return 1;
	}
}

/* There really should be a "htp_write_vtoc" in here.
 * But for some odd reason the kernel linker wont let us call
 * hpt_strategy() from here. So I had to put it in
 * hpt.c
 */

