/*
 * @(#) hpt-queues.c 1.7@(#)
 * queues.c
 *	This file holds routines to manage the "waiting" queues,
 *	and the "active" queues for each disk.
 *	The queues are linked lists hanging off the softstate info.
 *
 *	Only routines in this file should manipulate them directly.
 *
 *	Top part of the file has "wait queue" routines.
 *	Bottom part of the file has "active queue" routines
 */

#include <sys/errno.h>
/* THis stuff is for struct softstate, not because we need it here */

#include <sys/open.h>
#include <sys/stat.h>

#include <sys/vtoc.h>

#include <sys/dktp/fdisk.h> /* just for defs.h */

#include <sys/ddi.h>
#include <sys/sunddi.h>

#include "defs.h"
#include "hpt.h"
#include "hpt-udma.h"
#include "hpt-ide.h"



/* Call this only if you definately want something printed .
 * This prints the driver-internal linked list of queued requeusts.
 * If you dont own the waitqueue_mutex, the output may be odd.
 * Actually, if the driver is busy, you may CRASH THE SYSTEM, if you dont
 * hold the mutex.
 */
void hpt_printwaitqueue(struct hpt_softstate *softstate, int disk){
	int count=0;
#ifdef PRINTQUEUE_SHOWACTIVE
	int acount=0;
#endif
	struct diskhd *queue=&softstate->waitqueue[disk];
	struct buf *ent;

	ent=queue->av_forw;
	while(ent != NULL){
		count++;
		if(ent->b_flags & B_READ){
			DLPRINT2(0,"   queued: block %lld (read)", ent->b_lblkno);
		} else {
			DLPRINT2(0,"   queued: block %lld (write) ", ent->b_lblkno);
		}

		ent=ent->av_forw;
	}

}


/* Remove one (or more) jobs from the waitqueue that can be put together
 * as a single I/O request for the disk.
 * 
 * Right now, just pick the first one.
 *
 * You MUST HAVE the waitqueue mutex before calling this function. It is
 * just a convenience routine for hpt_one_from_waitqueue!!!
 */
static struct buf*pick_queued_job(struct hpt_softstate *softstate, int drive){
	struct diskhd *ioqueue=&softstate->waitqueue[drive];
	struct buf *chain=NULL;
	
	chain=ioqueue->av_forw;
	if(chain==NULL){
		DLPRINT(1,"WARNING: pick_queued_job called with empty queue");
		return NULL;
	}

	ioqueue->av_forw=chain->av_forw;
	chain->av_forw=NULL;

	/* Normally, we would now do
	* while(ioqueue->av_forw!=NULL){}
	* and try to coelese requests
	*/

	return chain;
	
}



/* return pointer to a job from the waitqueue for the drive,
 * removing it from the waitqueue in the process.
 * We grab and release the waitqueue mutex
 *
 */
struct buf *hpt_one_from_waitqueue(struct hpt_softstate *softstate, int drive){
	struct buf*retval;

	DLPRINT(2,"one_from_waitqueue: Getting ioqueue mutex");
	mutex_enter(&softstate->waitqueue_mutex[drive]);

	DLPRINT(2,"  Getting a job from wait queue");
	
	retval=pick_queued_job(softstate,drive);

	/* One way or another, we are now done with the wait queue.
	 * Either we've removed something already, or we failed and
	 * we're not going to do any more with it
	 */
	mutex_exit(&softstate->waitqueue_mutex[drive]);
	DLPRINT(2,"one_from_waitqueue: released ioqueue mutex, returning value");

	return retval;
}

/* routine for dequeue_all: EMERGENCY BAIL_OUT!!!
 * Specifically do NOT attempt to aquire mutex. ugh.
 */
void hpt_forceclean_waitqueue(struct hpt_softstate *softstate,int drive){
	struct buf*request;
	request=pick_queued_job(softstate,drive);
	while(request!=NULL){
		cleanup_iobuf(request,EIO);
		request=pick_queued_job(softstate,drive);
	}
}


/* Add an iorequest to the waiting queue for the drive, sorting
 * it in the process.
 *  (current sort field is bufptr->b_resid, supposedly)
 *    (we use disksort() )
 *
 * Aquires and releases the waitqueue_mutex, so the wait could be
 * rather long for this routine to finish.
 */
void hpt_add_to_waitqueue(struct hpt_softstate *softstate, int drive,
			  struct buf*newptr)
{

	struct diskhd *diskhdptr;

	if(newptr==NULL){
		DLPRINT(0,"hpt_add_to_waitqueue: ERRROR: NULL PTR!!\n");
		return;
	}
	diskhdptr=&softstate->waitqueue[drive];

	mutex_enter(&softstate->waitqueue_mutex[drive]);
	if(newptr->b_flags & B_READ){
		DLPRINT2(2,
			 "add_to_waitqueue: got queue mutex, inserting READ buf(%p)",
			 newptr);
	} else {
		DLPRINT2(2,
			 "add_to_waitqueue: got queue mutex, inserting WRITE buf(%p)",
			 newptr);
	}

	disksort(diskhdptr, newptr);


	if(debug_level>1){
		DLPRINT(0,"add_to_waitqueue: printing queue");
		hpt_printwaitqueue(softstate,drive);
	}
	mutex_exit(&softstate->waitqueue_mutex[drive]);
	DLPRINT(2,"add_to_waitqueue: insert done, releasing mutex");
}



/************************************************************/

/* This should only be called by the interrupt routine, or if you
 * are otherwise GUARANTEED the active queue will not change!!!
 */
void hpt_printactivequeue(struct hpt_softstate *softstate, int disk){
	struct diskhd *queue=&softstate->waitqueue[disk];
	struct buf *ent;
	ent=queue->b_forw;
	while(ent != NULL){
		if(ent->b_flags & B_READ)
			printf("   active: read block %lld",ent->b_lblkno);
		else
			printf("   active: WRITE block %lld",ent->b_lblkno);
		ent=ent->av_back;
	}
}



/* Set the activequeue to the given (sgclump *)
 *
 * Attempts to decrement the queue_sema for the BUS.
 * Therefore, it could be a while before this routine returns.
 * You probably dont want to call this from interrupt context!
 * 
 * return 0 if okay, 1 if something was already active
 */
int hpt_set_activequeue(struct hpt_softstate *softstate, int bus,
		    struct sgclump *newhead){
	int sema_return;
	DLPRINT2(1,"hpt_set_activequeue called with sgclump %p\n",newhead);
	if(bus>1){
		printf("ERROR: hpt_set_activequeue bus=%d\n",bus);
		return 1;
	}

	/* DO --NOT-- do "quick check" of queue before you have
	 * the semaphore!!!
	 */

	/* This is the ONLY PLACE sema_p is called for activequeue */
	sema_p(&softstate->activequeue_sema[bus]);

	sema_return=sema_tryp(&softstate->activequeue_sema[bus]);
	/* nasty paranoia thing, to see if we screwed up somewhere */
	if(sema_return!=0){
		DLPRINT2(0, "ERROR: hpt_setactive: could do DOUBLE sema_p(%d)!!",
			sema_return);
	}

	DLPRINT(1,"  hpt_set_activequeue: got activequeue SEMAPHORE.");

	if(softstate->activequeue[bus]!=NULL){
		printf("ERROR: hpt_set_activequeue: queue not NULL(0x%p, bus %d)\n",
		       softstate->activequeue[bus], bus);
		sema_v(&softstate->activequeue_sema[bus]);
		return 1;
	}

	softstate->activequeue[bus]=newhead;
	return 0;
}

/* This routine should ONLY be called from an interrupt routine
 * that holds the bus mutex and active-queue semaphore.
 * It's a half-peek into the mechanism, so should be
 * considered a "private" interface.
 *
 * It should be called to replace a failed request with a new one.
 * (This will probably only be used for my hack around broken
 *  DMA engine issues)
 * return 0 on okay, 1 if you screwed up and there was something
 * on the bus already.
 * Will NOT replace if something there already
 */
int hpt_replace_activequeue(struct hpt_softstate *softstate, int bus,
		    struct sgclump *newhead){
	if(bus>1){
		printf("ERROR: hpt_clear_activequeue bus=%d\n",bus);
		return NULL;
	}
	if(softstate->activequeue[bus]!=NULL){
		printf("ERROR: hpt_replace_activequeue: queue not NULL\n");
		return 1;
	}
	DLPRINT3(0,"hpt_replace_activequeue called: switching %p for %p",
	       newhead, softstate->activequeue[bus]);

	softstate->activequeue[bus]=newhead;
	return 0;
}

/*
 * assuming something is on the active queue, remove it.
 * return the pointer if there was something, or NULL otherwise.
 * You'd better be the interrupt routine, 'cause NOTHING ELSE
 * should be removing stuff from here.
 * 
 */
struct sgclump *hpt_clear_activequeue(struct hpt_softstate *softstate, int bus){
	struct sgclump *retval;
	if(bus>1){
		printf("ERROR: hpt_clear_activequeue bus=%d\n",bus);
		return NULL;
	}
	DLPRINT2(0,"hpt_clear_activequeue removing %p",
		 softstate->activequeue[bus]);
	retval=softstate->activequeue[bus];
	softstate->activequeue[bus]=NULL;
	return retval;
}


/* I HOPE I can get rid of this stuff */


/* a change from prior stuff. Only print "current" head of chain, and
 * anything chained off av_forward
 */
void printbufchain(struct buf *chain){
	struct buf *ent;

	if(chain==NULL){
		printf("  printbufchain passed NULL pointer?");
		return;
	}
	printf("printbufchain(buf %p)",chain);

	ent=chain;

	while(ent != NULL){
		if(ent->b_flags & B_READ){
			DLPRINT2(0,"   block %lld (read)", ent->b_lblkno);
		} else {
			DLPRINT2(0,"   block %lld (write) ", ent->b_lblkno);
		}

		ent=ent->av_forw;
	}
}




/* Convenience function for hpt_start_io.
 * return pointer to the "head" struct buf of the user request,
 * so it can more easily pick up read/write flag, and start block
 *
 * MAY RETURN NULL if you have screwed up and there is nothing active.
 */
struct buf* hpt_getactivehead(struct hpt_softstate*softstate,int drive){
	int bus=0;
	if(drive>1)bus=1;
	
	if(softstate->activequeue[bus]==NULL)
		return NULL;

	return(softstate->activequeue[bus]->bufptr);
}

/*
 * return pointer to the "head" struct sgclump of the user request,
 * which then points to the preallocated SGTABLE
 * You should only check stuff related to the sgtable, maek no other
 * assumptions.
 *
 * MAY RETURN NULL if you have screwed up and there is nothing active.
 */
struct sgclump* hpt_getactivesgclump(struct hpt_softstate*softstate,int drive){
	int bus=0;
	if(drive>1)bus=1;
	
	if(softstate->activequeue[bus]==NULL)
		return NULL;

	return(softstate->activequeue[bus]);
}

/* count valid number of blocks in active queue
 * Also ASSUMES that we can just total up the bytes because
 * we have reads of whole blocks.
 * The count will be a fairly restricted range, since the maximum IDE
 * transfer is only 128k or something like that.
 *
 * return -1 if NULL activequeue or bufptr somewhere .
 */
int hpt_countactiveblocks(struct hpt_softstate *softstate,int drive){
	struct buf *ent;
	struct sgclump *sgclumpptr;
	int bus=0;
	uint32_t bcount=0; /* BYTE COUNT */
	if(drive>1) bus=1;

	sgclumpptr=hpt_getactivesgclump(softstate,drive);

	if(sgclumpptr==NULL){
		printf("  hpt_countactive: NOTHING AcTIVE??");
		return -1;
	}


	while(sgclumpptr != NULL){
		ent=sgclumpptr->bufptr;
		if(ent==NULL){
			printf("ERROR: hpt_countactiveblocks found null bufptr");
			return -1;
		}
		bcount+= ent->b_bcount;

		sgclumpptr=sgclumpptr->next;
	}
	return (bcount >> DEV_BSHIFT);
}


/* Whether good or bad status, this will free up an sgclump chain.
 * (Usually found on the "active" queue)
 * if err=0, just call biodone on attached (struct buf*).
 * If err!=0, call bioerror as well
 *
 * called by hpt_cleanup_active, AND hpt_start_one_job
 *   (hpt_cleanup_active is normally itself called by hpt_handle_intr)
 *
 * MUTEXES:
 *   None. We ignore everything except the chain we are passed.
 *
 */
void hpt_free_sgclumpchain(struct sgclump *clumpptr,int err){
	struct sgclump *nextclump;
	if(clumpptr==NULL){
		printf("  hpt_free_sgclumpchain: nothing to clean? [CANNOTHAPPEN]\n");
		/* XXX should I call sema_v here? really, this
		 * clause should be a "cannot happen"
		 */
		return;
	}
	dealloc_sg_table(clumpptr);
	
	while(clumpptr!=NULL){
		struct buf*bptr;

		bptr=clumpptr->bufptr;
		if(bptr==NULL){
			printf("ERROR:  cleanup_active: sgclump with NULL bufptr!\n");
		} else {
			DLPRINT2(1,"cleanup_active cleaning buf %p",bptr);

			if(bptr->b_flags& B_PAGEIO){
				/* used for kernel allocated stuff */
				DLPRINT(1,"  bptr has B_PAGEIO set");
			}
			if(bptr->b_flags& B_PHYS){
				/* Normal for USERIO stuff */
				DLPRINT(1,"  bptr has B_PHYS set");
			}

			cleanup_iobuf(bptr,err);
		}
		nextclump=clumpptr->next;
		kmem_free(clumpptr,sizeof(struct sgclump));
		clumpptr=nextclump;
	}

}

/* hpt_cleanup_active()
 * Called when we have either successfully completed an interrupt,
 * or if an interrupt triggered with ERROR
 *
 * This takes whats on the "active" queue for the drive,
 * (which means the whole bus)
 * remove it, and free up everything associated with it.
 *
 *  [free up sgtable from sg_clump, free up sgclump,
 *   call biodone  on (struct *buf)
 *   and go down the sgclump->next chain]
 *
 * It is UP TO THE CALLER to ensure there are no multithread conflicts.
 *
 * If there is an error state, specify the errno you want returned to
 * the user in 'err'. Normally, use EIO.
 *
 * MUTEXES:
 * It is now UP TO THE CALLER to release the semaphore.
 *  (IFF we return 0)
 *
 *
 * CONTEXT: This may be called indirectly from INTERRUPT context,
 *  so it must not use stuff like cmn_err()
 *
 * returns 0 on okay, 1 if there was nothing on activequeue
 *
 */
int hpt_cleanup_active(struct hpt_softstate *softstate,int drive, int err){
	struct sgclump *clumpptr;
	int bus=drive>>1;

	if(err!=0){
		DLPRINT2(1,"WARNING: in hpt:cleanup_active, errno==%d\n",err);
	}

 	clumpptr=softstate->activequeue[bus];
	softstate->activequeue[bus]=NULL;
	DLPRINT3(1,"hpt_cleanup_active removed %p from queue %d",clumpptr,bus);

	if(clumpptr==NULL){
		/* even though hpt_free_sglumpchain is safe on NULL,
		 * DO NOT CONTINUE!
		 * Do NOT raise semaphore, if nothing processed!!!
		 */
		return 1;
	}

	hpt_free_sgclumpchain(clumpptr,err);

	return 0;
		
}


/*
 * called by hpt_handle_interrupt.
 * 
 * our goal:
 *  Unstick the UDMA hardware, somehow.
 *
 * One possible plan:
 * start current hardware with DMA job, that does nothing
 * but transfer sector 0 via DMA, to a throwaway buffer.
 * or SOMETHING, anyways.
 * Hopefully, this will unstick things
 */
void hpt_hackery(struct hpt_softstate *softstate,int drive){
	uint8_t command=IDECMD_IDENTIFY;
	uint8_t sectbuf[512];
	int bus=0;
	if(drive>1) bus=1;

	disable_interrupt(softstate,bus);


	do{
		if(ide_select_drive(softstate,drive) !=0){
			DLPRINT(0,"hpt_hackery FAILED in selecting drive");
			break;
		}
		/* nocheckstat, because currently, getpiodatablock reads status*/
	
		if(idecmd_nocheck(softstate,bus,command) !=0){
			DLPRINT(1,"hpt_hackery FAILED writing ide cmd");
			break;
		}

		if(getpiodatablock(softstate,bus,sectbuf) != 0){
			DLPRINT(0,"hpt_hackery FAILED on piodatablock");
			break;
		}
		DLPRINT(0,"hpt_hackery: successfully read info info");
	} while(0);
	enable_interrupt(softstate,bus);
	
#ifdef RESET_WITH_NOP
		DLPRINT(1,"  going for NOP to reset stuff. first disable intr");
		disable_interrupt(softstate,bus);
		DLPRINT(1,"  Now actual NOP, with double-read to clear");
		do_nop(softstate,bus,0);
		printstatus(clear_ide_err(softstate,bus));
		readaltstatus(softstate,bus); /* delay */
		printstatus(readstatus(softstate,bus));
		readaltstatus(softstate,bus); /* delay */
#endif
#ifdef RESET_WITH_SRESET
		if(software_reset(softstate,bus)!=0){
			DLPRINT(1,"software_reset FAILED?!!");
		} else {
			DLPRINT(1,"software_reset completed");
		}
#endif
#ifdef RESET_INIT_UDMA
		if(init_udma(softstate,disk)!=0){
			DLPRINT(1," INIT_UDMA FAILED");
		} else {
			DLPRINT(1,"INIT_UDMA returned okay");
		}
#endif
}
